#include <cxxtls/muSED.h>
#include <cxxtls/file.h>
#include <cxxtls/foreach.h>

#include <iostream>
#include <fstream>
#include <stdlib.h>

using namespace std;
using namespace cxxtls;

///@file mus
///  MicroSED executable wrapper for script execution.

///@page mus
///
///  <h1>MUS</h1>
///  mus is a program which invokes the @ref MicroSED ( cxxtls/muSED.h ) functionality in manner convenient
///  to script writers.  Generally, for scripting work, SED is to be prefered to
///  mus because SED is a more complete implementation of the SED language.  However,
///  @ref MicroSED scripts are not a pure subset of SED and so an executable that can
///  be invoked from bourne shell scripts, or other, is needed to complete the 
///  muSED toolset.
///
///  mus is basically like sed except that it does not support labels or any of the
///  commands that involve them and (at the time of this writing) does not support
///  the commands like "n" which force the reading of the next line of input.
///
///  mus (and cxxtls/muSED.h ) does implement an additional command, W, which stands for "while"
///  that lets the script repeatedly process the current line of input while a 
///  given regular expression remains true.   This is most helpful in the situation 
///  where the entire file has been concatenated into a single long line and then
///  you want to perform some sort of repeated processing on that line.
///
///  The 'O' command is also supported in both muSED and mus.  This command lets you 
///  sort the current contents of the pattern buffer, assuming that a newline delimited
///  list of lines.  There are various sort options.
///
///  Use the "-h" or "--help" command options to get help.
///
///
///

#ifdef __GNUC__
#   if __GNUC__ > 4 || ( __GNUC__ == 4 &&  __GNUC_MINOR__ > 2 )
#       define USE_muSED
#   endif
#else
#   define USE_muSED
#endif



#ifdef USE_muSED



string programName;

static void printHelp();

static void bomb(string const &msg)
{
   cerr << programName << " - fatal error:  " << msg << endl;
   exit(1);
}

static void apply(muSED::CompiledScript &script, list<string> const &container)
{
    list<string> output;

    muSED::apply(script, container, output);

    CXXTLS_FOREACH(string const &cur, output)
    {
       cout << cur << endl;
    }

}

static void readFile(string const &name, list<string> &container)
{
    if(name == "-")
    {
        string line;
        
        while(getline(cin, line))
        {
           container.push_back(line);
        }
    }
    else
    {
        string line;

        ifstream file(name.c_str());

        if(!file.good())
        {
           bomb(string(" reading input file ") + name );
        }
        
        while(getline(file, line))
        {
           container.push_back(line);
        }

    }
}

void applyToFile(muSED::CompiledScript &script,
                 std::string const     &filename,
                 bool                   lineByLine
                )
{
    if(lineByLine)
    {
        // process each line from the input file as whole file body
        // in and of itself.  This means that the hold commands
        // most likely will not work, though you can use them 
        // for complicated string operations.  Certainly they 
        // don't work if you try to hold the whole file into 
        // the hold buffer then swap it back into the pattern 
        // buffer and process it as one giant line.

        if(filename == "-")
        {
            string line;
            
            while(getline(cin, line))
            {
               list<string> fileContents;

               fileContents.push_back(line);

               apply(script, fileContents);

            }
        }
        else
        {
            string line;
            
            ifstream file(filename.c_str());
            
            if(!file.good())
            {
               bomb(string(" reading input file ") + filename );
            }
            
            while(getline(file, line))
            {
               list<string> container;
               container.push_back(line);

               apply(script, container);
            }
        }

    }
    else
    {
           list<string> fileContents;

           readFile(filename, fileContents);

           apply(script, fileContents);
    }
}


int main(int argc, char **argv)
{
   if(argc == 1)
   {
      printHelp();
      exit(1);
   }


   ////////////////////////////////////////////////////////////////////////////
   // consume options
   ////////////////////////////////////////////////////////////////////////////

   FileName  me(*argv++);

   programName = me.basename();

   list<string> script;  // -e and -f options get accumulated into this script

   bool addPrint = true; // -n option inverts this

   bool lineByLine = false; // -1 sets this

   for(; *argv; ++argv)
   {
       char const *optionPtr = *argv; 

       if(optionPtr[0] != '-' || optionPtr[1] == 0 )
       {
           break;
       }

       string option(optionPtr);

       if(option == "-f")
       {
           ++argv;

           if(*argv == 0)
           {
              bomb("expected script name after -f");
           }

          ifstream scriptFragment(*argv);

          if(!scriptFragment.good())
          {
              bomb(string("reading file ") + *argv);
          }

          string line;

          while(getline(scriptFragment, line))
          {
              script.push_back(line);
          }


       }
       else
       if(option == "-e")
       {
           ++argv;

           if(*argv == 0)
           {
              bomb("expected script string after -e");
           }
           
           script.push_back(*argv);
       }
       else
       if(option == "-1")
       {
          lineByLine = true;
       }
       else
       if(option == "-n")
       {
           addPrint = false;
       }
       else if(option == "-h" || option == "--help")
       {
          printHelp();
       }
       else
       {
          bomb(string("unknown option: ") + *argv);
       }
   }

   // when we get here, *argv will either be a null pointer or it will
   // point to the first input file.

   if(addPrint)
   {
       script.push_back("p");
   }


   ////////////////////////////////////////////////////////////////////////////
   // compile script
   ////////////////////////////////////////////////////////////////////////////

   muSED::CompiledScript executor(script);

   if(!executor.ok())
   {
       cerr << programName << " - fatal error in the following script" << endl;

       int line=1;

       CXXTLS_FOREACH(string const &cur, script)
       {
           cerr << "[" << line++ << "]=" << cur << endl;
       }

       cerr << endl;

       bomb(executor.error());
           
   }

   ////////////////////////////////////////////////////////////////////////////
   // execute the script on all incoming files
   ////////////////////////////////////////////////////////////////////////////

   if(*argv == 0)
   {
      // read from stdin and quit

      applyToFile(executor, "-", lineByLine);

   }
   else
   {
       while(*argv != 0)
       {
           FileName name(*argv++);

           applyToFile(executor, name, lineByLine);
       }

   }


    exit(0);
}

/////////////////////////////////////////////////////////////////////////////////////
static void printHelp()
{
   char const *lines[]=
   {
      "mus - MICRO-SED",
      "",
      "Usage:",
      "    mus  [options] [files]",
      "",
      "Description:",
      "",
      "   mus is a micro-sed.  It provides a proper subset of sed's capabilities",
      "   but also provides additional commands and features that sed does not.",
      "   mus is not preferred to sed, it exists to allow scripts consistent",
      "   with the muSED implementation to be used to perform textual transformation.",
      "",
      "Options",
      "",
      "    -f  scriptName    The name of file containing a mus script.",
      "    -e  scriptText    A string defining a script.",
      "    -n                Suppresses automatic printing of every line.",
      "    -h                Print this help message",
      "    -1                Line by line mode -- unbuffered processing.",
      "                      (Not compatible with 'q', 'Q', 'd'",
      "                       'D', or 'O' commands.  Hampers use of 'h',",
      "                       'H', 'g', and 'G' commands.)",
      "    --help            Print this help message",
      "",
      "Online Documentation",
      "   See http://www.bordoon.com/tools/index.html",
      "       mus",
      "       MicroSED",
      "       include/cxxtls/muSED.h",
      "",
      "Script Gramer",
      "",
      "   A mus script is a sequence of statements, each one of which has",
      "   an optional line-filtering component.  If the filter is left out",
      "   then the statement applies to all input lines, otherwise the",
      "   statement only applies to those which match the filter.",
      "",
      "   So, in BNF:",
      "",
      "      mus_statement ::=  [mus_line_filter] mus_statement",
      "",
      "   Statements are terminated either with a semicolon, ';', or",
      "   with the end of the line.  However, mus_block_statement is",
      "   allowed to span multiple lines, a mus_block_statement is defined",
      "   like this:",
      "",
      "      mus_block_statement := '{' { [newline] [statement] } '}'",
      "",
      "   A mus_block_statement is a collection of statements which all",
      "   share the same line-filtering criteria.",
      "",
      "   When specified, line-filters fall into one of the following groups",
      "",
      "    single line number   -- just a number that specifies a line",
      "                            number in the input stream, say '12'",
      "                            (Not compatible with -1 option)",
      "    end of line          -- the dollar sign, '$', means after the",
      "                            last line of the file has been handled",
      "                            perform this one last statement",
      "    /regex/              -- If the current pattern buffer matches",
      "                            the regex, then execute the statement",
      "                            on he pattern buffer.  The pattern",
      "                            buffer usually contains the current",
      "                            line of input, but statements can change",
      "                            it as they execute",
      "   <filtr1>,<filter2>    -- A pair of line filters that select a",
      "                            range of lines in the input stream on",
      "                            which to execute the statement.",
      "                            (Not fully compatible with -1 option)",
      "",
      "   Statement types:",
      "",
      "     s/match/replc/opts --  Substitute a given regex match for a",
      "                            specified replacement.  This is the most",
      "                            valuable command in mus and is described",
      "                            below.",
      "     p                  --  Print the current line to the output",
      "                            stream",
      "     P                  --  Print only up to the first newline found",
      "                            in the pattern buffer tot he output stream.",
      "     d                  --  Stop processing the current input line",
      "                            without printing it, but keep going to the",
      "                            next line of input.  (Not compatible with",
      "                            the -1 option.)",
      "     q                  --  Print current line of input and process",
      "                            no more lines of input.  (Not compatible",
      "                            with the -1 option.)",
      "     Q                  --  Do not print the current line of input and",
      "                            do not process any more lines.  (Not",
      "                            compatible with the -1 option.)",
      "     x                  --  Swap the pattern and hold buffers.  The",
      "                            pattern buffer is automatically populated",
      "                            with the next line of input it is processed.",
      "                            The hold buffer is a temporary variable that",
      "                            can hold the entire file, concatentated",
      "                            together using the h or H commands.",
      "     h                  --  Copy the pattern buffer to the hold buffer.",
      "                            (Not fully compatible with -1 option).",
      "     H                  --  Append the pattern buffer to the hold buffer",
      "                            using newline as a delimiter.",
      "                            (Not fully compatible with -1 option).",
      "     g                  --  Copy the hold buffer to the pattern buffer.",
      "                            (Not fully compatible with -1 option).",
      "     G                  --  Append the the hold buffer to the pattern",
      "                            buffer using newline as a delimiter.",
      "                            (Not fully compatible with -1 option).",
      "     y/set1/set2/       --  Translate characters in set1 into the",
      "                            corresponding character in set2.  The sets",
      "                            are defined using the same logic as the tr",
      "                            command uses -- and as regex character sets",
      "                            are defined.",
      "     O/options/         --  Treat the pattern buffer as a sequence of",
      "                            lines and sort them using the specifed",
      "                            options:",
      "                              u -- sort and discard duplicates",
      "                              r -- sort in reverse order",
      "                            (Not fully compatible with -1 option).",
      "                            Note that you will probably have wanted",
      "                            to append lines to the hold buffer, then",
      "                            swapped them back into the pattern buffer",
      "                            for this command to be useful.  Here is an",
      "                            example invocation that sorts a file and",
      "                            removes duplicate lines on Windows:",
      "                              mus -n -e \"H;${x;O/u/;p}\"",
      "                            On unix or linux use ' instead of \".",
      "",
      "Substitute Statement:",
      "",
      "  The substitute statement is how you replace string, but it is much",
      "  more than that.  A trivial substitution looks like this:",
      "",
      "    s/fred/bill/g",
      "",
      "  This command means:  replace all instances of fred with bill.  The",
      "  g character at the end means 'global.'  If you only wanted to replace",
      "  the first instance, you would use '1' instead of 'g'.  To use a case",
      "  insensitive replacement, add the 'i' option alongside the '1' or 'g'.",
      "",
      "  The power of the substitute command comes not from the simple",
      "  replacement, not even regular expression replacement, but rather from",
      "  it's ability to form replacement expression that are composed of parts",
      "  of the match.  For example, consider the following command:",
      "",
      "     s/\\([Ff]\\)red/\\1/g",
      "",
      "  Here, the word 'fred' would be converted to 'f' and if you had used",
      "  'Fred', it would be 'F'.",
      "",
      "  Making effective use of regular expression substitutions is documented",
      "  in many places on the net.",
      "",
      "",
      0
   };

   char const **scan = lines;

   while(*scan)
   {
      cout << *scan++ << endl;
   }

   exit(1);
}
#else
    int main() { cout << "MUS not supported on this OS" << endl;  return 0; }
#endif
