#ifndef ETAGS_H_INCLUDED
#define ETAGS_H_INCLUDED
//
// Copyright 2002, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure are covered by the
// same copyright.  Some of them are part of the GNU source distribution
// and you must obey the GPL copyright for those files.

///@file
///
///  This file defines the interface to the etags database -- the file
///  created when you run the etags program.

#include <string>
#include <portable_io.h>
#include <list>

namespace cxxtls
{

class EtagsDB
    ///  A class that encapsulates the searching of the database
    ///	 created by the	"etags"	program.  EtagsDB lets you read	the database
    ///  file and search it for a symbol. See "man etags" or a google search.
{
  FILE*       file_;      ///<  when file_ is 0, then the filename was bad
  std::string filename_;  ///<  name of etags file
  std::string symbol_;    ///<  name of symbol to find

public:

 EtagsDB(std::string etags_database_filename); ///< construct and open the specified file
 EtagsDB();                                    ///< default constructor

 ~EtagsDB();                                   ///< destructor

 bool ok() const { return file_ != 0; }

 bool set_tags_file(std::string tagsfile);  ///< Change the name of the tags
					    ///  database file -- closing the
					    ///  current one and opening a
					    ///  new one.  The function returns
					    ///  true if error.


 bool find_first(std::string const &symbol,
		 std::string       *file_name,
		 int               *line_number,
		 std::string       *line_text,    // text of line of definition
		 bool              case_sensitive_search=false
		);
   ///<
   ///  Find the first definition of a specified symbol name.  This function
   ///  must be called before find_next() -- which lets you find subsequent
   ///  references to the symbol.  The function return value indicates that
   ///  the symbol was in fact found. The third parameter, case_sensitive_search,
   ///  is used only if you want to find an exact match for the symbol including
   ///  case.
   ///
   ///  This search function assumes you are looking for C++ symbols and will
   ///  return class members matching the symbol name, namespace members matching
   ///  the symbol name, and global symbols -- including \#defines.
   ///
   ///  This function starts searching at the beginning of the database.


 bool find_next( std::string       *file_name,
		 int               *line_number,
		 std::string       *line_text
		);
    ///<
    ///  Find next is the same as find_first except that it does not allways
    ///  start at the beginning of the tags database when doing searches --
    ///  rather it starts at the location just passed the most recent find
    ///  first operation.


 int matching_files(std::string relative_path, std::list<std::string> *output) const;
     ///<
     ///  Find all files in the etags database that 'match' the specified
     ///  relative path.  
     ///
     ///  If the relatative path is actually an absolute path,
     ///  the returned list will contain exactly 1 element -- the input path --
     ///  regardless of whether or not it is in the etags database.
     ///
     ///  If the relative_path is a single file basename, or a subdirectory
     ///  path, such as "dir/dir/file", then all pathnames in the etags database
     ///  ending in "/" followed by the relative path will be returned.
     ///  Acount of matchines will be returned.
     ///
     ///  Note:  if the relative_path is empty string, then all files in the
     ///  etags database will be returned

		
private:

  bool case_sensitive_;

  EtagsDB(EtagsDB const &);               ///<  Copy constructor is private

  EtagsDB& operator= (EtagsDB const&);    ///<  Assignment operator is private

  char most_recent_file_[2048];           ///<  used in parsing


  int  searchFor(  std::string const &symbol,
		   std::string       *foundFile,
		   int               *foundLine,
		   std::string       *sourceText
		);
    ///<
    ///  This helper function actually scans the TAGS file.  
    ///
    ///  @return an integer containing one of the following:
    ///  @arg 0  didn't find the symbol but not end of file
    ///  @arg 1  did find the symbol
    ///  @arg 2  didn't find the symbol and end of file occurred
    ///  And 3 pointers to variables to hold additional return data
    ///
    ///
    ///  @param symbol defines the string to be searched for starting
    ///                at the current position in the TAGS database
    ///                file.
    ///  @param foundFile points to a variable which will be populated
    ///                   with the name of the file that defines the
    ///                   specified symbol.
    ///  @param foundLine points to a variable that will be updated with
    ///                   the line number within the foundFile where the
    ///                   symbol is defined.
    ///  @param sourceText points to a string variable that will be updated
    ///                    the text of the line that defines the symbol.

};

} // namespace cxxtls

#endif
