//
// Copyright 2012-2015 Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure are covered by the
// same copyright.  Some of them are part of the GNU source distribution
// and you must obey the GPL copyright for those files.

#include <cxxtls/stringexpression.h>
#include <cxxtls/stringexpressionparser.h>
#include <cxxtls/foreach.h>
#include <cxxtls/ScriptLanguageException.h>

#include <ctype.h>
#include <string.h>
#include <memory>

#include <stdlib.h>

using namespace std;

namespace cxxtls
{
static string apostrophe("'");
static string doubleQuote("\"");
static string slash("\\");


//============================================================================================
template<class Iterator1, class Iterator2>
Iterator1
findPathSeparator(Iterator1 f, Iterator2 l)
{
    while (f != l)
    {
        char c = *f;

        if(c == '/')
           break;

#ifdef _MSC_VER
        if(c == '\\')
           break;
#endif

        ++f;

    }

    return f;
}


//============================================================================================

    std::string StringExpression::stringify() const  // debug print helper
    {
         std::string rv;

         StringExpr::Expression::const_iterator f=expression_.begin(), l = expression_.end();

         while(f != l)
         {
             StringExpr::Expression::value_type const &cur = *f++;

             rv += cur->stringify();

             if(useSpaceSeparators_ && (f != l))
               rv += ' ';
         }

         return rv;
    }

//============================================================================================

StringExpression::~StringExpression()
{
  StringExpr::Expression::iterator f = expression_.begin(), l = expression_.end();
  
  while(f != l)
  {
      delete *f++;
  }

}

//============================================================================================

struct StringRegionInfo
{
   std::string rv_;
   std::vector<StringExpression::RegionDescriptor> regions_;
   size_t curSize_;
   size_t prevSize_;

   StringRegionInfo()
   :  curSize_(0)
   ,  prevSize_(0)
   {
   }
};


void StringExpression::executeHelper(ExecutionContext *context, StringArray &output, bool globbing, bool splitting) const
{
   return executeHelper_1(context, output, globbing, splitting, false);
}

std::string StringExpression::executeEncapsulated(ExecutionContext *context) const
{
   StringArray output;

   executeHelper_1(context, output, false, false, true);   

   return output.combine();
}

void StringExpression::executeHelper_1(ExecutionContext *context, StringArray &output, bool globbing, bool splitting, bool encapsulated) const
{
   

   // expand the string expression first by combinging the
   // components, including those components whose results are
   // created from variable expansions, "$var", or from 
   // complex expressions, "${var/source/dest}".
   //
   // Finally, perform glob expansion:  *.c => a.c b.c etc

   // all output begins an expression that evaluates to a single string.
   // that string has some regions that are "globabable" and some that are
   // not.   
   //
   // Once the initial string is created, the globbable parts are, well globbed,
   // and the results are sent to the final output as an array -- probably
   // with only one value, in the case of most strings which are non-globable,
   // but some are globable.


   std::vector<StringRegionInfo> regionInfo;

   regionInfo.push_back(StringRegionInfo());



   StringExpr::Expression::const_iterator f = expression_.begin(), 
                                          l = expression_.end();


   // some regions of the resultant string are protected from
   // glob expansion, so we need to keep track of the start
   // and end of those regions.

   regionInfo.back().prevSize_=0;
   regionInfo.back().curSize_ = 0;

   std::list< std::pair< bool, std:: string > > pendingResults;  

   int chunkCount= 0;

   bool allowsNoResult = false;  // only "$@" allows no result

   while(f != l || !pendingResults.empty())
   {
       bool protection;

       if(pendingResults.empty())
       {
           // this is the normal case -- normally, expressions only produce a single output
           
           StringArray tmpOutput;  // evaluate each sub expression into it's
                             // own temporary variable

           if(encapsulated && !isRedir_)
           {
               protection = true;
               tmpOutput.push_back( (*f++)->executeEncapsulated(context) );
           }
           else
           {
               if ((*f)->allowsNoResult())
                   allowsNoResult = true;  // probably should be region specific, but lets try it like this

               protection = (*f++)->executeHelper(context,tmpOutput, splitting);
           }

           if (!protection && splitting)
           {
               std::vector< std::string > output;

               for (size_t i = 0; i < tmpOutput.size(); ++i)
                   context->splitWords(tmpOutput[i], output);

               tmpOutput.clear();

               CXXTLS_FOREACH(std::string const &cur, output)
                   tmpOutput.push_back(cur);
           }

           size_t tmpSize = tmpOutput.size();

           if(!tmpSize)
           {
              // nothing to add to rv
           }
           else
           {
               if (useSpaceSeparators_)
                   regionInfo.back().rv_ += ' ';

               ++chunkCount;

               // then concatenate the results

               regionInfo.back().rv_ += tmpOutput[0]; 

               if(tmpSize > 1)
               {
                  for(size_t i = 1; i < tmpSize; ++i)
                      pendingResults.push_back( std::pair<bool, std::string> (protection, tmpOutput[i]) );
               }
           }
       }
       else
       {
           // this is the case for "$@" wherein an expression can result in multiple values

           regionInfo.push_back(StringRegionInfo());  // treat all the members of the multiple output
                                                      // group as new returnable string

           regionInfo.back().prevSize_ = 0;
           regionInfo.back().curSize_  = 0;

           regionInfo.back().rv_ = pendingResults.front().second;
           protection            = pendingResults.front().first;

           pendingResults.pop_front();

       }




       // now handle the protection status of this new text.

       regionInfo.back().curSize_ = regionInfo.back().rv_.size();

       if(regionInfo.back().curSize_ != regionInfo.back().prevSize_)
       {
          if(protection)
          {
             if(!regionInfo.back().regions_.empty() && regionInfo.back().regions_.back().first)
             {
                // this region is protected and so was the
                // most recent region -- so let's just
                // widen the previous region's protection
                // zone.

                regionInfo.back().regions_.back().second.second = regionInfo.back().curSize_;
             }
             else
             {
                 // this new region is protected and the
                 // previous region either did not exist
                 // or was not protected, so let's add the
                 // new protection region:

                  regionInfo.back().regions_.push_back( RegionDescriptor(protection, 
                                              CharacterRange(regionInfo.back().prevSize_, 
                                                             regionInfo.back().curSize_)
                                                     ) 
                                   );
             }
          }
          else
          {
              // if this region is not protected from globbing,
              // let's check to see if the region is split by one
              // or more /'s.

              std::string::const_iterator f = regionInfo.back().rv_.begin() + regionInfo.back().prevSize_,
                                          l = regionInfo.back().rv_.begin() + regionInfo.back().curSize_;

              while(f != l)
              {
                  std::string::const_iterator where = findPathSeparator(f,l);
                  
                  size_t slashOffset = (where - regionInfo.back().rv_.begin());


                  
                  if(slashOffset < regionInfo.back().curSize_)
                  {
                    // we found a / delimited region -- check to see if has * or ? in it
                    // if not, then we mark this region as protected for 
                    // later globbing to reduce unnecessary calls to
                    // to FileName::find_matching_names which are expensive.


                     // found a / between regionInfo.back().prevSize_ and regionInfo.back().curSize_
                  
                    if(regionInfo.back().prevSize_ != slashOffset)
                    {
                         bool prot = (where == std::find_if(f, where, StringExpr::IsGlobable()));

                         regionInfo.back().regions_.push_back( RegionDescriptor(prot, 
                                                 CharacterRange(regionInfo.back().prevSize_, 
                                                                slashOffset)
                                                     ) 
                                          );

                    }
                     // now tack on the trailing / as a region of it's won

                     regionInfo.back().regions_.push_back( RegionDescriptor(false, 
                                                 CharacterRange(slashOffset, 
                                                                slashOffset+1)
                                                     ) 
                                   );

                     regionInfo.back().prevSize_ = slashOffset + 1;
                     f = where + 1;

                  }
                  else
                     break;
                  
              }

              if(regionInfo.back().prevSize_ != regionInfo.back().curSize_)
              {
                     regionInfo.back().regions_.push_back( RegionDescriptor(protection, 
                                                 CharacterRange(regionInfo.back().prevSize_, 
                                                             regionInfo.back().curSize_)
                                                     ) 
                                   );
              }

          }

       }

       if(regionInfo.back().curSize_)
       {
          if(regionInfo.back().regions_.empty() || regionInfo.back().regions_.back().second.second != regionInfo.back().curSize_)
             throw cxxtls::ScriptLanguage::Exception("StringExpression::executeHelper", "reason", "execution halted before end of expression");
                            // string is not empty, the end of the range
                            // should be equal to regionInfo.back().curSize_

          if(regionInfo.back().regions_.empty() || regionInfo.back().regions_[0].second.first != 0)
          {
             throw cxxtls::ScriptLanguage::Exception("StringExpression::executeHelper", "reason", "expression corrupt");
          }
       }

       regionInfo.back().prevSize_= regionInfo.back().curSize_;
   }

   if(!regionInfo.back().rv_.empty())
   {
      if(regionInfo.back().regions_.empty())
      {
          throw cxxtls::ScriptLanguage::Exception("StringExpression::executeHelper", "reason", "expression corrupt");  // we should have at least 1 region
      }

      if(regionInfo.back().regions_[0].second.first != 0)
      {
         throw cxxtls::ScriptLanguage::Exception("StringExpression::executeHelper", "reason", "expression corrupt");  // the first region should be 0
      }

      if(regionInfo.back().regions_.back().second.second != regionInfo.back().rv_.size())
      {
         throw cxxtls::ScriptLanguage::Exception("StringExpression::executeHelper", "reason", "expression corrupt");  // the first region should be 0
      }

   }

  
   CXXTLS_FOREACH(StringRegionInfo &cur, regionInfo)
   {
       if (allowsNoResult && cur.rv_.empty() )
          continue;  // for "$@" skip truly empty results, everyone else gets a row in the output even for empty strings.

      executeHelperGlobHelper(context, output, globbing, cur.rv_, cur.regions_);
   }

}
//======================================================================================
//======================================================================================
//======================================================================================
void 
StringExpression::
executeHelperGlobHelper(ExecutionContext *context, StringArray &output, bool glob, std::string const &rv, std::vector<RegionDescriptor> &regions) const
{

   // At this point, rv is properly populated and the regions array describes
   // sections of it -- some are protected from globbing.  Some of it are
   // not.  Of those regions that are not protected, there are no /'s in that
   // region unless it is the first character of the region.  This makes the
   // globbing logic easier.


   if(!glob || !context->globbing())
   {
      // glob is state as defined in the execution logic of the string expression.
      //
      //  context->globbing() is the user's preference of whether we should be globbing at all

      output.push_back(rv);
   }
   else
   {  // now glob!


      StringArray result;   // tmp variable containing globbed version of result



       std::vector<RegionDescriptor>::iterator f = regions.begin(),
                                               l = regions.end();


       // [f,l) represents a collection of contiguous regions in
       // rv.  These regions may or may not have globbing characters
       // in them (and if they do they may not be in a globbable region,
       // as defined by f->first.

       string filePattern;
       bool shouldGlob = false;

       while(f != l)
       {  // construct pattern for fnmatch which is called by find_matching(), below

         // in each potentially glob-able region, get a character pointer range
         // in rv into [firstByte,lastByte)

           std::string::const_iterator firstByte = rv.begin() + f->second.first,
                                       lastByte  = rv.begin() + f->second.second;

           if(f->first)
           {
              // current region is protected

              while(firstByte != lastByte)
              {
                 char c = *firstByte++;

                 if(c == '*' || c == '?')
                 {
                     filePattern += '[';
                     filePattern += c  ;
                     filePattern += ']';
                 }
                 else
                 if(c == '\\')
                    filePattern += '/';
                 else
                     filePattern += c;
              }

           }
           else
           {
              // possibly globable

              while(firstByte != lastByte)
              {
                 char c = *firstByte++;

                 if(c == '*' || c == '?')
                 {
                     shouldGlob = true;
                     filePattern += c;
                 }
                 else
                 if(c == '\\')
                    filePattern += '/';
                 else
                     filePattern += c;
              }

           }

           ++f;

       }  // construct pattern for fnmatch


       if(shouldGlob)
       {
           list<string> matches;
           
           preGlobFilter(filePattern, context);   // convert leading ~/s into references to the home directory

           FileName::glob(filePattern.c_str(), matches);
           
           if(!matches.empty())
           {
               list<string>::iterator a = matches.begin(),
                                      b = matches.end();
           
               while(a != b)  
               {  // iterate over files found by glob

                    std::string tmp = *a++;

                    #ifdef _MSC_VER
                         char sepchar = '\\';

                         size_t count = tmp.size();

                         std::string converted;

                         converted.reserve(tmp.size());

                         CXXTLS_FOREACH(char c, tmp)
                         {
                            if(c == '/')
                               c = sepchar;

                            converted.push_back(c);
                         }

                         tmp = converted;


                    #else
                         char sepchar = '\\';

                    #endif

           
                    if(!tmp.empty())
                    {
                        if(tmp[0] == '.' &&  tmp.size() > 1 && tmp[1] == sepchar)
                        {
                            result.push_back(std::string(tmp, 2, tmp.size()-2));   // remove ./ from local directory name
                        }
                        else
                            result.push_back(tmp); // if no ./, then juse use the local directory name.  should not get here, but who knows
           
                    }
                    else
                         result.push_back(tmp);  // handle empty strings from glob -- huh?

               } // iterate over files found by glob
           }
           else
               result.push_back(rv);   // count == 0.  Count of matching files from find_matching()
       }
       else
       {
           // !shouldGlob

           string tmp(rv);

           preGlobFilter(tmp, context);   // convert leading ~'s into references to the home directory, if needed

           result.push_back(tmp);
       }


       for(size_t i = 0; i < result.size(); ++i)
           output.push_back(result[i]);

   } // now glob!

}


//======================================================================================
//======================================================================================
//======================================================================================
//======================================================================================
//======================================================================================



namespace StringExpr
   /// implementation details for the StringExpression class.
{

///////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////
void ExecutionContext::findVariableOrValue(std::string const &name, std::string &rv) const 
{ 
   char c;

   static std::string empty;

   if(name.size() == 1  && !isalpha(name[0]))
   {
       if(isdigit(c = name[0]))
       {
           if(c == '0')
           {
               rv = scriptName();
           }
           else
           {
                size_t index = c - '1';

                if(arguments_.size() > index)
                   rv = arguments_[index];
                else
                   rv= empty;
           }
       }
       else 
       if(c == '#')
       {
          // count of arguments, not including argument 0

          char buffer[40];

          snprintf(buffer, 40, "%d", int(arguments_.size()-1));

          rv = buffer;
       }
       else 
       if(c == '?')
       {
          char buffer[40];

          snprintf(buffer, 40, "%d", int(errorCode_) );

          rv = buffer;
       }
       else
       {
          rv = empty;
       }

   }
   else
       rv = find(name); 
}

///////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////

std::string VariableExpansionNode::executeEncapsulated(ExecutionContext *context) const
{
   StringArray output;

   executeHelper(context, output, false);

   return encapsulateLiteralDoubleQuotes(output.combine());
}
///////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////

bool VariableExpansionNode::executeHelper(ExecutionContext *context, StringArray &output, bool splitting) const 
{

   output.push_back(std::string());

   string &rv = output.back();

   // This function handles the argument vector as well as simple variables.
   //
   // To wit:  a vector of strings is the input to various expansion algorithms.
   // If we are working with a variable expansion, then we are working on a single
   // element vector.  If we are working on the argument vector, we work on a
   // collection of separate argument values and concatenate the results at the
   // end (with space separators).
   // 

   std::vector<std::string> workVector;

   if(name_.empty())
   {
      // how did we get here anyway?

      rv.resize(0);

      return false;
   }

   // determine the contents of the work vector, the input data to the expansion
   // algorithms.

   bool singleVariable = true;    // we are processing some variable name, not $* or $@

   switch(name_[0])
   {
       case '*':
       case '@':
           {
               // we are working on all the arguments to this script

               if(length_)
               {
                  char buffer[40];

                  snprintf(buffer, sizeof(buffer), "%lu", context->arguments().size()-1); // don't include 0 in the list

                  rv = buffer;

                  return false;
               }

               if(name_[0] == '@')
                   singleVariable = false;  // we want to keep the values as separate array elements

               std::vector<std::string>::const_iterator f = context->arguments().begin(),
                                                        l = context->arguments().end();

               if(f != l)
                  ++f;       // first array element doesn't count

               if (f == l && output.size() == 1)
               {
                   output.clear();  // if this is an empty $@, then we don't want to return an empty string if there are no args, we want to return NOTHING
               }

               while(f != l)
               {
                  std::string const &cur = *f++;

                  workVector.push_back(cur);

               }
           }
           break;
       case '0': case '1':  case '2': case '3': case '4': case '5': case '6': case '7': case '8': case '9':
           {
               size_t index = name_[0] - '0';

               if(index  < context->arguments().size())
               {
                  workVector.push_back(context->arguments()[index]);
               }
           }
       break;

       case '#':
          {
              // count of arguments
              workVector.push_back(lexical_cast<std::string>(context->arguments().size() -1 ));
          }
          break;

       default:

           // we are working on a single variable expanded

           if(name_ == "_")
           {
               workVector.push_back(context->lastParmOfLastCommand());
              
           }
           else
           {

               context->findVariableOrValue(name_, rv);
               
               workVector.push_back(rv);
           }

           break;

   }

  // at this point, singleVariable and work are populated correctly based
  // on whether or not we are operating on $*/$@ or a named variable expansion
  // such as $3 or $fred.

   std::vector<std::string>::iterator wvFirst = workVector.begin(),
                                      wvLast  = workVector.end()
                                      ;
   while(wvFirst != wvLast)
   {
        std::string &work = *wvFirst++;

        if(substring_)
        {
           // the substr_ flag means we are not doing a replacement, but rather
           // that we are extracting a substring.
           //
           // that is, we are modifying the work variable so that it contains
           // a substring of itself.

           StringArray output;

           if (pattern_.get())
               pattern_->executeHelper(context, output, false, splitting);

           std::string matchString = output.combine(); 
        
           int offset=-1;

           int length=1000 * 1000 * 1000;  // a billion should cover "all" of the string length

           sscanf(matchString.c_str(), "%d", &offset);

           if(offset < 0 || offset >= work.size())
               work = string();

           {
               // length is optional
               output.clear();

               replacement_->executeHelper(context, output, false, splitting);

               matchString = output.combine(); 

               if (!matchString.empty())
               {
                   int tmp=-1;

                   sscanf(matchString.c_str(), "%d", &tmp);

                   length = tmp;
               }
           }

           if(length < 1)
               work = string();

           if(offset < work.size())
           {
              // we have not indexed off the end of the string

              if(offset + length > work.size())
                 length = work.size() - offset;  

              work = work.substr(offset, length);
           }
        }
        else
        if(length_)
        {
           char buffer[40];
        
           if(singleVariable)
           {
                snprintf(buffer, sizeof(buffer), "%lu", work.size());
           }
           else
           {
                // we should not get here, but for safety sake, I've written code
                // that should work.
                snprintf(buffer, sizeof(buffer), "%lu", workVector.size());
           }
        
           work = buffer;

           if(!singleVariable)
             
            {
              rv = work;
              return false;  // return the length of the  args vector (should never get here)
             }

        }
        else
        if(upperCase_)
        {
            if(!work.empty())
            {
        
                char c;
        
                if(upperCase_ == 1)
                {
                    // uppercase only the first character
        
                    c = work[0];
        
                    if(islower(c))
                    {
                        work.replace(0, 1, 1, toupper(c));
                    }
                }
                else
                {
                    // uppercase all the characters that need it.
        
                    std::string tmp;
        
                    char const *f = work.data();
                    char const *l = f + work.size();
        
                    while(f !=l)
                    {
                       c = *f++;
        
                       if(islower(c))
                          c = toupper(c);
        
                       tmp += c;
                    }
        
                    work = tmp;
        
        
                }
            }
        }
        else
        if(lowerCase_)
        {
            if(!work.empty())
            {
                char c;
        
                if(lowerCase_ == 1)
                {
                    // uppercase only the first character
        
                    c = work[0];
        
                    if(isupper(c))
                    {
                        work.replace(0, 1, 1, tolower(c));
                    }
                }
                else
                {
                    // uppercase all the characters that need it.
        
                    std::string tmp;
        
                    char const *f = work.data();
                    char const *l = f + work.size();
        
                    while(f !=l)
                    {
                       c = *f++;
        
                       if(isupper(c))
                          c = tolower(c);
        
                       tmp += c;
                    }
        
                    work = tmp;
        
        
                }
            }
        }
        else
        if(first_)
        {
           // remove leading instances of pattern from work
        
              StringArray output;

              pattern_->executeHelper(context, output, false, splitting);

              std::string matchString = output.combine(); 
        
              PatternMatcher matcher(matchString);
        
              work = matcher.removeLeadingMatch(work, all_);
        
        }
        else
        if(last_)
        {
           // remove trailing instances of pattern from work

              StringArray output;
        
              pattern_->executeHelper(context, output, false, splitting);
              
              std::string matchString = output.combine();
        
              PatternMatcher matcher(matchString);
        
              work = matcher.removeTrailingMatch(work, all_);  
        
        }
        else
        if(pattern_.get())
        {
           StringArray output;
           
           pattern_->executeHelper(context, output, false, splitting);

           std::string matchString = output.combine();
        
           // replace matches of the pattern_ with the replacement
        
           // all_ means replace all matches, else
           // beginOnly_ means replace a match only at the beginning
           // endOnly_ means replace a match only at the end
           // else replace first match in the string
        
           if(!matchString.empty())
           {
        
               PatternMatcher matcher(matchString);
        

               StringArray output;
               replacement_->executeHelper(context, output, false, splitting);

               std::string repTemp = output.combine();
               
               std::string replacement;

               // decode escaped values

               char const * f = repTemp.data();
               char const * l = f + repTemp.size();

               while(f != l)
               {
                  char c = *f++;

                  if(c != '\\')
                     replacement += c;
                  else
                  {
                     if(f == l)
                        replacement += '\\';
                     else
                        switch(c = *f++)
                        {
                            default:
                                replacement += c;
                                break;

                            case 'n':  replacement += '\n'; break;
                            case 't':  replacement += '\t'; break;
                            case 'r':  replacement += '\r'; break;
                            case 'e':  replacement += 0x1b; break;
                            case 's':  replacement += ' ';  break;
                            case 'b':  replacement += 0x08; break;
                            case 'a':  replacement += 0x07; break;  // alarm, bell
                        }
                  }

               }

               // perform the replacement

        
        
               if(all_)
               {
                   // replace all
        
                   work = matcher.replaceMatches(work, replacement, true);
        
               }
               else
               if(beginOnly_)
               {
        
                   size_t matchLength = matcher.longestLeadingMatch(work);
        
                   if(matchLength)
                   {
                       std::string tmp = replacement;
        
                       tmp.append(work.begin() + matchLength,
                                  work.end()
                                 );
        
                       work = tmp;
                   }
        
               }
               else
               if(endOnly_)
               {
                   size_t matchLength = matcher.longestTrailingMatch(work);
        
                   if(matchLength)
                   {
                       std::string tmp = work.substr(0, work.size() - matchLength);
        
                       work = tmp + replacement;
                   }
               }
               else
               {
                  // replace the first instance of the match, no matter
                  // where it is in the string
        
                   work = matcher.replaceMatches(work, replacement, false);
        
               }
           }
        
        }
   }

    // join the workVector elements into a space delimited string
   
   if (!singleVariable && !workVector.empty())
   {
       // we are returning an array, not a single variable

       output.pop_back();   // remove the 'rv' we added

       CXXTLS_FOREACH(std::string const &cur, workVector)
           output.push_back(cur);

       return false;
   }
   else
   {

      // rv is suspect because "$@" of an empty args vector gives you an empty output
      // not a single line in the output.
      //
      // rv is a reference variable pointing to a deleted line in that case.

      if(output.empty() && workVector.empty())
         return false;

      std::string *returnValue;

      if(output.empty())
      {
         output.push_back(std::string());
      }

      {
          string &tmp = output.back();  

          returnValue = &tmp;
      }


   
    (*returnValue).resize(0);

    std::vector<std::string>::iterator f = workVector.begin(), l = workVector.end();

    while(f != l)
    {
       (*returnValue) += *f++;

       if(f != l && !(*returnValue).empty())
          (*returnValue) += context->sepChar();
    }

    }

    return false; 
}

///////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////

static std::string  defaultSystemCommandExecutor(std::string const &command,
                                                 void*              threadExecutionContext // ignored
                                                )
{

   string rv;

   FILE *f = popen(command.c_str(), "r");

   if(f)
   {
       for(;;)
       {
          int c = fgetc(f);

          if(c < 0)
             break;

          rv += c;
       }
   }

   size_t offset = rv.find_last_not_of("\r\n");

   if (offset < rv.size())
   {

     rv = rv.substr(0, 1 + offset);

   }


   return rv;

}

///////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////

StringExpr::ExecutionContext::SystemCommandExecutorFunc
StringExpr::ExecutionContext::systemCommandExecutor_ =    &defaultSystemCommandExecutor;


///////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////

std::string SubShellNode::executeEncapsulated(ExecutionContext *context) const
{
    StringArray output;

    executeHelper(context, output, false);

    if(encodeResults_)
    {
        string rv("\"");

        string tmp = output.combine();

        string::const_iterator  f = tmp.begin(),
                                l = tmp.end();
        while(f != l)
        {
           char c = *f++;

           switch(c)
           {
               default:     rv += c;
                            break;

               case '"':
               case '\\':   rv += '\\';
                            rv += c;
                            break;

           }
        }  // while not out of input data

        rv += '"';
        return rv;
    }
    
    return output.combine();

}

///////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////

bool  SubShellNode::executeHelper(ExecutionContext *context, StringArray &output, bool splitting) const
{ 


   output.push_back(std::string());

   string &rv = output.back();
   
   string command;

   bool firstWord = true;

   list<Node*>::iterator f = command_->begin(),
                         l = command_->end();
   while(f != l)
   {

       Node *p = *f++;

       if(!command.empty())
          command += ' ';

       if (firstWord)
       {
           firstWord = false;
           StringArray tmp;

           p->executeHelper(context,tmp,false);

           string result = tmp.combine();
           
           command += result;

           char c = 0;

           if (!result.empty())
              c = result[0];


           if(c == ';' || c == '|' || c == '&')
              firstWord = true;
       }
       else
       {
           string tmp = p->executeEncapsulated(context);

           command += tmp;

           char c = 0;

           if (!tmp.empty())
              c = tmp[0];

           if(c == ';' || c == '|' || c == '&')
              firstWord = true;
       }

   }

   rv = context->executeSystemCommand(command);


   return !splitResults_;

}

///////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////


std::string VariableExpansionNode::stringify() const
{ 
    std::string rv("${");

    rv += name_;

    if(first_)
    {
       rv += '#';

       if(all_)
         rv += '#';

       rv += pattern_->stringify();

    }
    else
    if(last_)
    {
       rv += '%';
       if(all_)
          rv += '%';

       rv += pattern_->stringify();

    }
    else
    if(pattern_.get())
    {


       rv += '/';
       if(all_)
          rv += '/';

       rv += pattern_->stringify();

       rv += '/';

       rv += replacement_->stringify();

    }


    rv += '}';

    return rv;
    
}  // debug print helper


///////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////


void
ExecutionVariables::
getProcessVariables(ExecutionVariables::ProcessVariableSet &output) const
{
#ifdef _MSC_VER

    output.vars_.clear();

    if(vars_.empty())
       return;  // nothing to do

    // to speed things up, compute needed size and reserve said space in the output vector

    size_t neededSize = 4;  // the trailing 4 bytes that ends the vars object for CreateProcess


    CXXTLS_FOREACH(VariablesMap::value_type const &cur, vars_)
    {
       neededSize += 2;   // for = and trailing nul

       neededSize += cur.first.size();   // name
       neededSize += cur.second.size();  // value
    }

    output.vars_.reserve(neededSize);

    // now copy the data to the output

    CXXTLS_FOREACH(VariablesMap::value_type const &cur, vars_)
    {
        if(!cur.second.exportFlag())
           continue;  // only produce exported values

        CXXTLS_FOREACH(char c, cur.first)  output.vars_.push_back(c);

        output.vars_.push_back('=');

        CXXTLS_FOREACH(char c, cur.second)  output.vars_.push_back(c);

        output.vars_.push_back(char(0));
    }

    output.vars_.push_back(char(0));
    output.vars_.push_back(char(0));
    output.vars_.push_back(char(0));
    output.vars_.push_back(char(0));




#else
   
#endif


}  // getProcessVariables()

///////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////

void ExecutionVariables::includeProgramEnvironment()
{
#ifdef _MSC_VER
    char **env = _environ;
#else
    char **env = environ;
#endif

    while(*env)
    {
        char *setting = *env++;

        char *equal = strchr(setting, '=');

        if(!equal)
           continue;

        std::string varname, varvalue;

        varname.append(setting, setting + (equal - setting));

#ifdef _MSC_VER
        if(   varname.size() == 4 
           &&
              ( varname[0] == 'p' || varname[0] == 'P' )
           && 0 == strcmpi(varname.c_str(), "PATH") 
          )
        {
           varname="PATH";
        }
#endif


        varvalue = (equal+1);  // and the rest of the value

        set(varname, varvalue);
        setExportFlag(varname,true);

    }

    char buffer[2001];

    ::getcwd(buffer, sizeof(buffer)-1);

    buffer[sizeof(buffer)-1] = 0;

    set("PWD", buffer);



}


///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
   void
   ExecutionVariables::
   addToExecHash(std::string const &commandName, std::string const &fullProgramPath, int hashEntityType)
   {


       execHash_[commandName] = std::pair<std::string, int> (fullProgramPath, hashEntityType);
   }
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////

    std::pair<std::string,int>
    ExecutionVariables::
    findInExecHash(std::string const &programName, int hashEntityType)
    {

       FileName progPath = programName;

       // the input program name is not an aboslute path.

       std::map<std::string, std::pair<std::string,int> >::iterator where = execHash_.find(programName);

       if(where != execHash_.end())
       {
          FileName tmp(where->second.first);

          if(tmp.exists() && !tmp.is_dir())
             return where->second;

          // old hash location may be current directory relative path and may no longer be in
          // existence in the current directory

          execHash_.erase(where);  // we will pretend that we did not see it
       }

       // We do not have a previous use of this programName

       FileName tmp = FileName::find_executable(programName, get("PATH"));

       if(!tmp.empty())
       {
          // we were able to find the specified programName in the search.

          if(tmp.is_executable())
          {
             execHash_[programName] = std::pair<std::string, int>(tmp, 0);    // FileName.find_executable() only returns native executables
             return std::pair<std::string,int>(tmp, 0);

          }
          else
              return std::pair<std::string, int>(tmp, -1);   // don't know what it is, but we did find it
       }


       return std::pair<std::string,int>(tmp,-1);  // we did not find it

    }  // ExecutionVariables::findInExecHash()


ExecutionContext::UndefinedVariablesHandlers *
ExecutionContext::undefinedVariablesHandlers_=0;   // assumes not to ever change and must

void
ExecutionContext::
WordSplitter::
setIFS(std::string const &ifs)
{
    ifs_ = ifs;

    if(ifs_.find_first_not_of(" \t\n\r") < ifs_.size())
    {
       wordSplitting_ = false;
    }
    else
    {
       wordSplitting_ = true;
    }

}  // WordSplitter::setIFS


static size_t max_splits = 1000 * 1000 * 1000;

void
ExecutionContext::
WordSplitter::
split(std::string const &input, std::vector< std::string > &output) const
{

   if(wordSplitting_)
   {
       StrTool::stringlist_t outlist;

       StrTool::parse_words(input, &outlist, max_splits, *this);

       CXXTLS_FOREACH(std::string const &cur, outlist)
         output.push_back(cur);

   }
   else
   {
      StrTool::splitOnDelimiter(input, output, *this);

   }


} // ExecutionContext::WordSplitter::split()


void ExecutionContext::setLastParmOfLastCommand(std::string const &value)
{
    lastParm_.resize(0);
    lastParm_.append(value.begin(), value.end());
}

int ExecutionContext::getTmpDirName(std::string &result)
{

#ifdef _MSC_VER
        result = find("TEMP");

        if(result.empty())
        {
            result = find("USERPROFILE");

        }

        FileName tmpFile(result + string("\\AppData\\Local\\Temp"));

        if(tmpFile.is_dir())
        {
           result = tmpFile;
        }

#else
        result = find("TMP");

        if(result.empty())
           result = "/tmp";
        
#endif

        FileName tmp(result);

        if(!tmp.is_dir())
        {
           return -2;
        }

        return 0;

}

std::string ExecutionContext::quoteChunk(std::string const &s)
    // quote a string fragment
{
    if(s.find_first_of("'*?") < s.size())
    {
       // use double quotes

       string rv;

       rv.reserve(s.size() + 2);

       rv += '"';

       bool bugWorkaround = false;   // right now, bug exists in glob code.  You can glob this:  "stuff"*.c, but you can't glob *".c".  I'll fix later

       string::const_iterator f = s.begin(),
                              l = s.end();
       while(f !=l)
       {
           char c = *f++;

           if(c == '\\')
           {
              rv += c;          // escape the escape
              rv += c;

              if( f != l )
              {
                 rv += *f++;   // and leave the thing after the escape unmolested
              }
           }
           else
           if(c == '"')
           {
              rv += '\"';
              rv += c;
           }
           else
           if(c == '*' || c == '?')
           {
               rv += '"';
               rv += c;
               bugWorkaround = true;
           }
           else
              rv += c;
       }

       if(! bugWorkaround)
           rv += '"';


       return rv;

    }
    else
    {
       // otherwise, just use apostrophes

        if(s.find_first_of(" \t\n{}()") < s.size())
            return apostrophe + s + apostrophe;

        return s;
    }
}

std::string SubExpressionNode::executeEncapsulated(ExecutionContext *context) const
{ 
    return value_->executeEncapsulated(context);
}



}   // StringExpr

StringExpression::PreGlobFilterFunction StringExpression::preGlobFilter_ = 0;  // default is a do-nothing function

} // cxxtls
